/*  ----------------------------------------------------------------------

    Copyright (C) 2001  Tom Vandenhove  (tom.vandenhove@ubizen.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    ---------------------------------------------------------------------- */

#include "nomadJB_data.h"
#include "data.h"
#include "callbacks.h"
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

extern int debug;
extern struct songlist *songs_root;
extern struct genre *genres_root;
extern struct artist *artists_root;
extern struct albumlist *albums_root;
extern struct playlist *playlist_root;
extern struct selection selection;

char *make_message(const char *fmt, ...) 
{
	/* Guess we need no more than 100 bytes. */
	int n, size = 100;
	char *p;
	va_list ap;
	if ((p = malloc (size)) == NULL)
		return NULL;
	while (1) {
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf (p, size, fmt, ap);
		va_end(ap);
		/* If that worked, return the string. */
		if (n > -1 && n < size)
			return p;
		/* Else try again with more space. */
		if (n > -1)    /* glibc 2.1 */
			size = n+1; /* precisely what is needed */
		else           /* glibc 2.0 */
			size *= 2;  /* twice the old size */
		if ((p = realloc (p, size)) == NULL)
			return NULL;
	}
}

char *get_homedir(void)
{
	struct passwd *pw;
        int uid;
        char *homedir;
        char *username;
        
        homedir = getenv ("HOME");    
        /* 
	 * first we call getlogin. 
	 * There might be several accounts sharing one uid 
	 */
	if ( homedir )
		return homedir;
        
        pw = 0;
        
        username = getenv("LOGNAME");
        if ( !username )
		username = getlogin();
        if ( username )
	pw = getpwnam( username);
	
        if ( pw == 0 ){
		/* if we can't getlogin, look up the pwent by uid */
		uid = geteuid();
                pw = getpwuid(uid);
        }
	/* we might still get no entry */
	if ( pw )
                return pw->pw_dir;
        return 0;
}

/* auxiliary for checklist */
static int albumpresent(struct albumlist *albums, struct album *album)
{
	while ( albums != NULL ) {
		if ( strcmp(albums->album->name, album->name) == 0 )
			return 1;
		albums = albums->next;
	}
	return 0;
}

/* checks if the the artist (or genre) for this song (and album) is already
   listed. If yes, add the album to the artists album list. If not, create 
   a new artist (genre) and add the album there */
static void checklist(struct artist **p_pointer, char *name, 
		      struct album *album, SONG *song)
{
	struct artist *pointer = *p_pointer;
	struct artist *prev_pointer = *p_pointer;
	while ( pointer != NULL ) {
		if ( strcmp(pointer->name,name) == 0 ) {
			struct albumlist *tmp,*prev;
			if ( albumpresent(pointer->albums, album) == 0 ) {
				tmp = pointer->albums;
				while ( tmp != NULL ) {
					prev = tmp;
					tmp = tmp->next;
				}
				tmp = (struct albumlist *)malloc(sizeof(struct albumlist));
				tmp->album = album;
				tmp->next = NULL;
				if ( pointer->albums == NULL ) pointer->albums = tmp;
				else prev->next = tmp;
			}
			break;
		}
		prev_pointer = pointer;
		pointer = pointer->next;
	}
	if ( pointer == NULL ) {
		pointer = (struct artist *)malloc(sizeof(struct artist));
		pointer->name = name;
		pointer->albums = (struct albumlist *)malloc(sizeof(struct albumlist));
		pointer->albums->album = album;
		pointer->albums->next = NULL;
		pointer->next = NULL;
		if ( prev_pointer == NULL ) *p_pointer = pointer;
		else prev_pointer->next = pointer;
	}
}

void add_song_to_list(struct songlist **songlist, SONG *song)
{
	*songlist = (struct songlist *)malloc(sizeof(struct songlist));
	(*songlist)->song = song;
	(*songlist)->next = NULL;
}

void add_song_to_end_of_list(struct songlist **root, SONG *song)
{
	struct songlist *tmp = *root;
	if ( *root == NULL ) add_song_to_list(root,song);
	else {
		while (tmp->next) tmp = tmp->next;
		add_song_to_list(&tmp->next,song);
	}
}

/* adds song to correct album, creating it if necessary.
   returns album song was added to */
struct albumlist *add_song_to_albums(struct albumlist **root,SONG *song)
{
	struct albumlist *prev_album = NULL;
	struct albumlist *album = *root;
	while ( album != NULL ) {
		if ( strcmp(album->album->name,song->album) == 0 ) {
			add_song_to_end_of_list(&album->album->songs,song);
			break;
		}
		prev_album = album;
		album = album->next;
	}
	if ( album == NULL ) {
		album = (struct albumlist *)malloc(sizeof(struct albumlist));
		album->album = (struct album *)malloc(sizeof(struct album));
		album->album->name = song->album;
		add_song_to_list(&album->album->songs,song);
		album->next = NULL;
		if ( *root == NULL ) *root = album;
		else prev_album->next = album;
	}
	return album;
}

/* sorts all structures */
void sort_all()
{
	struct artist *artist;
	struct genre *genre;

	sort_albums(&albums_root);
	sort_artists(&artists_root);
	sort_artists((struct artist **)&genres_root);

	artist = artists_root;
	while ( artist != NULL ) {
		sort_albums(&artist->albums);
		artist = artist->next;
	}
	
	genre = genres_root;
	while ( genre != NULL ) {
		sort_albums(&genre->albums);
		genre = genre->next;
	}
}

/* adds a song to songlist, appropriate album list, and the 
   album to the appropriate artist/genre */
void add_song(SONG *song)
{
	struct albumlist *album;
	struct songlist *songs = songs_root;
	
	while (songs) {
		if (songs->song->idx == song->idx) {
			printf("song already present (idx %i match) - not adding to lists\n",(int)song->idx);
			return;
		}
		songs = songs->next;
	}
	printf("adding song %s\n",song->title);
	add_song_to_end_of_list(&songs_root,song);
	album = add_song_to_albums(&albums_root,song);
	checklist(&artists_root, song->artist, album->album, song);
	checklist((struct artist **)&genres_root, song->genre, album->album, song);
	sort_all();
}

void create_indexes(struct songlist *songs_root, 
		    struct albumlist **albums_root, 
		    struct artist **artists_root,
		    struct genre **genres_root)
{
	struct songlist *songs;
	struct artist *artist;
	struct genre *genre;
	TAG *tags;
	
	songs = songs_root;
	while ( songs != NULL ) {
		struct albumlist *album;
		if (debug) { printf("location : %li\n",songs->song->idx);}
		album = add_song_to_albums(albums_root,songs->song);
		checklist(artists_root, songs->song->artist, album->album, songs->song);
		checklist((struct artist **)genres_root, songs->song->genre, album->album, songs->song);
		songs = songs->next;
	}

	sort_all();
}

void read_lists()
{
  printf("reading lists\n");
  songs_root = (struct songlist *)nomadJB_filelist();
  playlist_root = (struct playlist *)nomadJB_playlist();
  //songs_root = NULL;
  //playlist_root = NULL;
}

void restore_lists()
{
        int libcount,idx, size;
	FILE *file;
	struct songlist **psongs;
	struct playlist **pplaylist;
	struct model_info *model_info;
	char *home,*path;

	model_info = nomadJB_model();
	home = get_homedir();
	path = make_message("%s/.nomadjb/cache.njb.%s",home,model_info->id);
	if ( ! ( file = fopen(path,"r") ) ) {
	  printf("file not opened \n");
	  read_lists();
	  return;
	}
	else {
	  printf("file opened \n");
	}
	
	if ( ! fscanf(file,"LIBCOUNT %i\n",&libcount) ) {
	  printf("libcount not read \n");
	  read_lists();
	  return;
	}
	
	if ( libcount != nomadJB_get_libcount() ) {
	  printf("libcount doesn't match\n");
	  read_lists();
	  return;
	}
	
	printf("libcount = %i matches... restoring from cache\n",libcount);

	/* restoring songs */
	songs_root = NULL;
	psongs = &songs_root;
	while ( fscanf(file,"SONG %i\n", &idx) ) {
		TAG *tags;
		TAG **oldtag;
		int type,data;
		char *label,*string;
		//printf("song %i\n",idx);
		oldtag = &tags;
		while ( fscanf(file,"TAG %i %i ", &type,&size) ) {
			TAG *newtag;
			newtag = (TAG *)malloc(sizeof(TAG));
			//printf("tag %i %i\n", type,size);
			label = (char *)malloc(size+1);
			fgets(label,size+1,file);
			//printf("tag %i %i\n", type,size);
			//printf("tag %i %i %s ", type,size,label);
			newtag->type = type;
			newtag->label = label;
			newtag->next = NULL;
			
			if ( type == 1 ) {
				fscanf(file,"%i",&data);
				//printf("%i\n",data);
				newtag->data = data;
			}
			else if ( type == 0 ) {
				fscanf(file,"%i ",&size);
				//printf("size %i\n",size);
				string = (char *)malloc(size+1);
				fgets(string,size+1,file);
				//printf("%i %s\n",size,string);
				newtag->string = string;
			}
			fgetc(file);
			*oldtag = newtag;
			oldtag = &newtag->next;
		}
		add_song_from_tags(psongs, idx, tags);
		psongs = &(*psongs)->next;
	}
	
	playlist_root = NULL;
	pplaylist = &playlist_root;
	/* restoring playlists */
	while ( fscanf(file,"PLAYLIST %i ", &size) == 1 ) {
		struct playlist *playlist;
		struct songlist **psongs;
		playlist = (struct playlist *)malloc(sizeof(struct playlist));
		playlist->name = (char *)malloc(size+1);
		fgets(playlist->name,size+1,file);
		//printf("playlist %s\n",playlist->name);
		fgetc(file);
		playlist->next = NULL;
		playlist->songs = NULL;
		psongs = &playlist->songs;
		while ( fscanf(file,"SONG %i\n", &idx) == 1 ) {
			struct songlist *songs;
			//printf("idx %i\n",idx);
			*psongs = (struct songlist *)malloc(sizeof(struct songlist));
			songs = songs_root;
			while ( songs ) {
				if ( idx == songs->song->idx )
					(*psongs)->song = songs->song;
				//printf("song %s\n",songs->song->title);
				songs = songs->next;
			}
			(*psongs)->next = NULL;
			psongs = &(*psongs)->next;
		}
		*pplaylist = playlist;
		pplaylist = &playlist->next;
	}
	
	printf("restore ended\n");
	
	//songs_root = NULL;
	//playlist_root = NULL;
	fclose(file);
	return;
}  

void save_lists()
{
	int libcount;
	struct songlist *songs;
	struct playlist *playlists;
	char *home,*path;
	FILE *file;
	struct model_info *model_info;

        libcount = nomadJB_get_libcount();
	printf("libcount %i\n",libcount);
	model_info = nomadJB_model();
	
	home = get_homedir();
	//path = (char *)malloc(sprintf("%s/.nomadjb",home)+1);
	path = make_message("%s/.nomadjb",home);
	//printf("home %s %s\n",home,path);
	mkdir(path,S_IRUSR|S_IWUSR|S_IXUSR);
	//path = (char *)malloc(sprintf("%s/.nomadjb/test.njb",home)+1);
	path = make_message("%s/.nomadjb/cache.njb.%s",home,model_info->id);
	//printf("home %s %s\n",home,path);
	creat(path,S_IRUSR|S_IWUSR);
	file = fopen(path,"w");
	free(path);
	/* free rest */
	
	fprintf(file,"LIBCOUNT %i\n",libcount);
	songs = songs_root;
	while (songs) {
		TAG *tag;
		fprintf(file,"SONG %i\n",songs->song->idx);
		tag = songs->song->tags;
		while (tag) {
			//printf("%i %s ",tag->type,tag->label);
			fprintf(file,"TAG %i %i %s ",tag->type,strlen(tag->label),tag->label);
			if ( tag->type == 0 ) {
				fprintf(file,"%i %s\n",strlen(tag->string),tag->string);
			}
			else {
				fprintf(file,"%i\n",tag->data);
			}
			tag = tag->next;
		}
		songs = songs->next;
	}
	
	playlists = playlist_root;
	while (playlists) {
		fprintf(file,"PLAYLIST %i %s\n",strlen(playlists->name),playlists->name);
		songs = playlists->songs;
		while (songs) {
			fprintf(file,"SONG %i\n",songs->song->idx);
			songs = songs->next;
		}
		playlists = playlists->next;
	}
	
	fclose(file);
}

void print_songs(struct songlist *songs) 
{
	while (songs) {
		SONG *song = songs->song;
		printf("song title  %s\n",song->title);
		printf("     artist %s\n",song->artist);
		printf("     album  %s\n",song->album);
		printf("     genre  %s\n",song->genre);
		printf("     year   %i\n",song->year);
		printf("     nr     %i\n",song->track_nr);
		printf("     length %i\n",song->length);
		printf("     size   %i\n",song->size);
		songs = songs->next;
	}
}

int album_compare(void const *a, void  const *b)
{
	//struct albumlist *album1, *album2;
	//int result;
	//album = *a;
	//printf("starting compare\n");
	//printf("name 1 %s\n",(*((struct albumlist **)a))->album->name);
	//printf("name 2 %s\n",(*((struct albumlist **)b))->album->name);
	return strcmp((*((struct albumlist **)a))->album->name, (*((struct albumlist **)b))->album->name);
	//printf("end compare\n");
	//return result;
	//return 1;
}

void sort_albums(struct albumlist **root)
{
	struct albumlist **array;
	struct albumlist *albums = *root;
	int size = 0;
	int i;
	//printf("sorting \n");
	while (albums) {
		size++;
		albums = albums->next;
	}
	array = (struct albumlist*)malloc(sizeof(struct albumlist*)*size);
	//printf("sizes %i %i\n",sizeof(struct albumlist)*size,sizeof(struct albumlist*)*size);
	albums = *root;
	size = 0;
	while (albums) {
		array[size] = albums;
		//printf("array %i %s\n",size,(struct albumlist*)array[size]->album->name);
		size++;
		albums = albums->next;
	}
	//for (i=0; i<size; i++) { printf("array %i %s\n",i,(struct albumlist*)array[i]->album->name); }
	qsort(array,size,sizeof(struct albumlist*),album_compare);
	//for (i=0; i<size; i++) { printf("array %i %s\n",i,(struct albumlist*)array[i]->album->name); }
	for (i=0; i<size-1; i++) { 
		//printf("array %i %s\n",i,(struct albumlist*)array[i]->album->name);
		(struct albumlist*)array[i]->next = (struct albumlist*)array[i+1];
	}
	//printf("sorting allmost done\n");
	(struct albumlist*)array[size-1]->next = NULL;
	*root = (struct albumlist*)array[0];

	free(array);
}

int artist_compare(void const *a, void  const *b)
{
	return strcmp((*((struct artist **)a))->name, (*((struct artist **)b))->name);
}

sort_artists(struct artist **root)
{
	struct artist **array;
	struct artist *artists = *root;
	int i,size = 0;
	while (artists) { size++; artists = artists->next; }
	array = (struct artist*)malloc(sizeof(struct artist*)*size);
	artists = *root;
	i = 0; while (artists) { array[i] = artists; i++; artists = artists->next; }
	qsort(array,size,sizeof(struct artist*),artist_compare);
	for (i=0; i<size-1; i++) (struct artist*)array[i]->next = (struct artist*)array[i+1];
	(struct artist*)array[size-1]->next = NULL;
	*root = (struct artist*)array[0];
	free(array);
}
